/*	$Id: ipc.c,v 1.19 2006-07-28 22:43:09 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/


// ipc stuff for NT 
// adapted from Michael Callahan's Win 95 VxD
// pjb

#define KERNEL 1

#include "errno.h"
#include "coda_types.h"
#include "coda.h"
#include "coda_nt.h"
#include "ipc.h"

// semaphore to signal a venus waiting in FS_IOCTL  during a get request
static KSEMAPHORE pending_msg;
static int pending_count;

/* list of requests waiting to be picked up by venus */
static struct dllist_head pending = {&pending, &pending};

/* list of requests currently being processed by venus */
static struct dllist_head processing = {&processing, &processing};

/* mutex to control access to the lists */
static ERESOURCE mtx; 


NTSTATUS request_init()
{
    static int needs_init=1;
    int rc = STATUS_SUCCESS;
    
    if ( needs_init == 1 ) {
	if ( ! list_empty(&processing) ) {
	    CDEBUG((D_IPC, "request_init: reinit of &processing\n"));
	    list_init(&processing);
	}
	if ( !list_empty(&pending) ) {
	    CDEBUG((D_IPC, "request_init: reinit of &pending\n"));
	    list_init(&pending);
	}
	// ExInitializeResourceLite can only return NT_SUCCESS?
	rc = ExInitializeResourceLite( &mtx );
	if ( rc != STATUS_SUCCESS ) {
	    CDEBUG((D_IPC, "request_init: error initializing mtx\n"));
	}
	pending_count = 0;
	KeInitializeSemaphore (&pending_msg, 0, 1000);
	needs_init = 0;
    }
    
    return rc;
}



static struct request *request_dequeue (struct dllist_head *list)
{
    struct dllist_head *le;
    struct request *req;
    if (list_empty(list))
	return NULL;
    
    le = list->prev;
    list_del(le);
    req = list_entry(le, struct request, le);
    
    return (struct request *) le;
}


static void request_remove (struct request *req)
{
    list_del(&req->le);
}

static struct request *dequeue_by_id (struct dllist_head *list,
				      unsigned long unique)
{
    struct dllist_head *le;
    struct request *req;
    
    for (le = list->next; le != list; le = le->next) {
	req = list_entry(le, struct request, le);
	if (req->req_data.in->ih.unique == unique)
	    break;
    }
    
    if (le == list)
	return NULL;   /* not there */
    list_del(le);
    
    return req;
}

// process_rpc: called by a thread making a request to Venus
// the thread will block until the call returns or times out
// unless timeout happens this doesn't remove the request
static unsigned long unique=1;

int request_process( struct request *req )
{
    int rc;
    int maxwait;
    KEVENT waitfor_venus;
    ERESOURCE_THREAD me;
    LARGE_INTEGER timeout;
    KIRQL irql;

    // Start with a short time out
    timeout = RtlConvertLongToLargeInteger(-SHORT_CODA_TIMEOUT);
    
    // request isn't done yet
    KeInitializeEvent( &waitfor_venus, SynchronizationEvent, FALSE );
    req->doneflag = &waitfor_venus;
    
    // add request to the pending list
    
    // since the wait parameter is TRUE, ignore return value
    me = ExGetCurrentResourceThread();
    ExAcquireResourceExclusiveLite( &mtx, TRUE );
    
    CDEBUG((D_IPC, "request_process: adding to pending unique %d,"
	    " pending_count %d\n", unique, pending_count));

    req->req_state = VM_QUEUED;
    req->req_data.in->ih.unique = unique;
    unique++;	
    list_add(&req->le, pending.prev);
    pending_count++;
    
    ExReleaseResourceForThreadLite( &mtx, me );

    CDEBUG((D_IPC | D_DETAILS, "request_process: Ready to Wait\n"));

#if DBG
    irql = KeGetCurrentIrql();
    if (irql != PASSIVE_LEVEL) {
	if (irql == DISPATCH_LEVEL)
	    DbgPrint("Calling KeWaitForSingleObject at DISPATCH_LEVEL\n");
	else if (irql == APC_LEVEL)
	    DbgPrint("Calling KeWaitForSingleObject at APC_LEVEL\n");
	else
	    DbgPrint("Calling KeWaitForSingleObject at ?_LEVEL\n");
    }
#endif

    //  Now signal the venus thread
    KeReleaseSemaphore ( &pending_msg, 0, 1, 1);

    // wait for reply 
    maxwait = 3;  // one short, two long timeouts
    rc = STATUS_TIMEOUT;  // to start the loop

    while (maxwait-- && (rc == STATUS_TIMEOUT)) {
	rc = KeWaitForSingleObject( &waitfor_venus, WAIT_FOR_VENUS,
				    KernelMode, TRUE, &timeout);
	
	CDEBUG((D_IPC | D_DETAILS, "request_process: After wait, rc=%d\n",
		rc));

	if (rc == STATUS_TIMEOUT) {
	    ExAcquireResourceExclusiveLite( &mtx, TRUE );
	    if (req->req_state ==  VM_QUEUED) {
		DbgPrint ("process_request returning timeout: unique %d, "
			  "pending_count %d, opcode %d\n",
			  req->req_data.in->ih.unique, pending_count,
			  req->req_data.in->ih.opcode);
		list_del(&req->le);
		pending_count--;
		KeClearEvent(&waitfor_venus);
		ExReleaseResourceForThreadLite( &mtx, me );
		// XXX Should shutdown everything here??
		return rc;
	    }
		
	    if (maxwait == 0) {
		DbgPrint ("process_request too many timeouts:  unique %d, "
			  "pending_count %d, opcode %d\n",
			  req->req_data.in->ih.unique, pending_count,
			  req->req_data.in->ih.opcode);
		list_del( &req->le );
		pending_count--;
	    } else
		timeout = RtlConvertLongToLargeInteger(-LONG_CODA_TIMEOUT);
	    
	    ExReleaseResourceForThreadLite( &mtx, me );
	}
    }
	
    // Clear the event in any case ...
    KeClearEvent(&waitfor_venus);

    CDEBUG((D_IPC, "request_process: Returning %d\n", rc));
    
    // back to life, let's tell the caller what happened
    return rc;
}


// similar to process_rpc but doesn't block

int insert_upcall( struct request *req )
{
    int rc;
    ERESOURCE_THREAD me;
    
    // request has no wait event
    req->doneflag = NULL;
    
    // add request to the pending list
    
    // since the wait parameter is TRUE, ignore return value
    me = ExGetCurrentResourceThread();
    ExAcquireResourceExclusiveLite( &mtx, TRUE );
    req->req_data.in->ih.unique = unique++;
    list_add(&req->le, &pending);
    pending_count++;
    ExReleaseResourceForThreadLite( &mtx, me );
    
    return 0;
}

/* Venus calls this on FETCH. This waits for pending_msg */
struct request *request_get_pending(int blockflag)
{
    DWORD postval, res;
    ERESOURCE_THREAD me;
    struct request *req;
    NTSTATUS rc;
    
    if (blockflag) {
	// 30 second relative wait.  Use to allow venus to exit.
	// Threads in DeviceIoControl wait apper to not be killable.
	// Returning once every 30 seconds should not be too much.

	LONGLONG timeout = -300000000; 

	// wait for pending to be non-empty
	rc = KeWaitForSingleObject(&pending_msg, WAIT_FOR_VENUS, KernelMode,
				   FALSE, (LARGE_INTEGER*)&timeout);
	if (rc != STATUS_SUCCESS) {
	    CDEBUG((D_IPC | D_DETAILS, "request_get_pending: timeout!\n"));
	    return 0;
	}
    }
    
    // add request to the pending list
    me = ExGetCurrentResourceThread();
    ExAcquireResourceExclusiveLite( &mtx , TRUE );
    req= request_dequeue(&pending);
    if ( req ) {
	pending_count--;
	list_add(&req->le, &processing);
    }
    ExReleaseResourceForThreadLite( &mtx , me);
    
    if (req)
	CDEBUG((D_IPC, "request_get_pending: %d: got request "
	       "(opcode %x). %d pending\n", 
	       req->req_data.in->ih.unique, 
	       req->req_data.in->ih.opcode,
	       pending_count));
    else
	CDEBUG((D_IPC, "request_get_pending: no request!\n"));
    return req;
}

/* find_request -- called by venus upon ANSWER to insert outbuf */
struct request *request_find (long unique)
{
    struct request *req;
    ERESOURCE_THREAD me;
    
    me = ExGetCurrentResourceThread();
    ExAcquireResourceExclusiveLite( &mtx, TRUE );
    req = dequeue_by_id (&processing, unique);
    ExReleaseResourceForThreadLite( &mtx, me );
    
    
    if (req)
	CDEBUG((D_IPC, "request_find: request %d: ready to answer\n",
		req->req_data.in->ih.unique));
    else
	CDEBUG((D_IPC, "request_find: request %d: ERROR--no such unique\n",
		unique));
    return req;
}


/* signal_request -- called by potemkin after copying reply back in */
void request_signal (struct request *req)
{
    // signal the event, don't wait 
    KeSetEvent(req->doneflag, 0, FALSE);
}


/* buffer size is enough for a full struct request plus two full filenames
   (for rename)
*/

/* allocate_request -- allocate a request structure */
struct request *request_allocate(void)
{
    struct request *req;
    
    req = (struct request *) ExAllocatePool (NonPagedPool, sizeof(*req));
    if (req) {
	RtlZeroMemory (req, sizeof(*req));
	list_init(&req->le);
    }
    return req;
}

void *request_allocate_data(struct request *req, int len)
{
    void *d;
    
    d = ExAllocatePool(NonPagedPool, len);
    if (d) {
	RtlZeroMemory (d, len);
    }
    req->req_bufflen = len;
    req->req_data.in = d;
    return d;
}

/* free_request -- free a request structure */
void request_free (struct request *req)
{
    ExFreePool (req);
}

void request_free_data (struct request *req)
{
    ExFreePool (req->req_data.in);
    req->req_data.in= NULL;
    req->req_bufflen = 0;
}


static void clear_queue_by_rh (struct dllist_head *list, DWORD rh)
{
    struct dllist_head *le, *le2;
    struct request *req;
    
    le = list->next;
    while (le != list) {
	req = list_entry(le, struct request, le);
	if ( req->rh == rh) {
	    le = le->next;
	    list_del(&req->le);
	    
	    req->req_data.out->oh.result = EIO;
	    request_signal (req);
	} else {
	    le = le->next;
	}
    }
}



static void clear_resource_requests (DWORD rh)
{
    ERESOURCE_THREAD me;

    me = ExGetCurrentResourceThread();
    ExAcquireResourceExclusiveLite( &mtx, TRUE );
    clear_queue_by_rh (&pending, rh);
    clear_queue_by_rh (&processing, rh);
    ExReleaseResourceForThreadLite( &mtx, me );
}

static void age_queue (struct dllist_head *list, DWORD agelimit)
{
    struct dllist_head *le, *le2;
    struct request *req;
    
    le = list->next;
    while (le != list) {
	req = list_entry(le, struct request, le);
	if (++req->age >= (int) agelimit) {
	    le = le->next;
	    list_del(&req->le);
	    req->req_data.out->oh.result = EIO;
	    request_signal (req);
	} else {
	    le = le->next;
	}
    }
}

